--------------------------------------------------------------------------------
-- Stand Alone Uplink: FPGA Design
-- LIGO document: D060288
-- SubModule TX_decoder
-- Created   9/15/2006 8:33:50 PM
--------------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use IEEE.Numeric_std.all;

entity TX_decoder is port
   (
     clk           : in    std_logic;
     reset         : in    std_logic;
     status        : out   std_logic_vector(15 downto 0);

     A_go          : in    std_logic;
     A_first       : in    std_logic;
     A_last        : in    std_logic;
     A_flush       : in    std_logic;
     B_go          : in    std_logic;
     B_first       : in    std_logic;
     B_last        : in    std_logic;
     B_flush       : in    std_logic;
     C_go          : in    std_logic;
     C_first       : in    std_logic;
     C_last        : in    std_logic;
     C_flush       : in    std_logic;
     D_go          : in    std_logic;
     D_first       : in    std_logic;
     D_last        : in    std_logic;
     D_flush       : in    std_logic;

     SEL           : out   std_logic_vector(1 downto 0);
     A             : out   std_logic_vector(6 downto 0);
     MC            : in    std_logic_vector(31 downto 0);

     transmit      : out   std_logic;
     enable        : out   std_logic;
     eof           : out   std_logic
   );
end TX_decoder;
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
architecture Structure of TX_decoder is

-- Type Declarations
   type tx_decoder_type is (IDLE, GO);

-- Signal Declarations
   signal reshist     : std_logic_vector (2 downto 0);
   signal longreset   : std_logic;

   signal A_golatch   : std_logic;
   signal A_ack       : std_logic;
   signal B_golatch   : std_logic;
   signal B_ack       : std_logic;
   signal C_golatch   : std_logic;
   signal C_ack       : std_logic;
   signal D_golatch   : std_logic;
   signal D_ack       : std_logic;

   signal first       : std_logic;
   signal last        : std_logic;
   signal middle      : std_logic;
   signal flush       : std_logic;

   signal state       : tx_decoder_type;
   signal qsel        : std_logic_vector (1 downto 0);
   signal addr        : unsigned (6 downto 0);
   signal eofprog     : std_logic;
   signal veto        : std_logic;
   signal decode_err  : std_logic;
   signal status_1    : std_logic_vector(15 downto 0);
   signal status_2    : std_logic_vector(15 downto 0);

begin
   -- prolong reset
   reshist <= reshist(1 downto 0) & reset when rising_edge(clk);
   longreset <= reshist(2) or reshist(1) or reshist(0);

   -- latch go flag
   process (A_go, A_ack, A_flush) is
   begin
      if (A_ack = '1') or (A_flush = '1') then
         A_golatch <= '0';
      elsif A_go = '1' then
         A_golatch <= '1';
      end if;
   end process;
   process (B_go, B_ack, B_flush) is
   begin
      if (B_ack = '1') or (B_flush = '1') then
         B_golatch <= '0';
      elsif B_go = '1' then
         B_golatch <= '1';
      end if;
   end process;
   process (C_go, C_ack, C_flush) is
   begin
      if (C_ack = '1') or (C_flush = '1') then
         C_golatch <= '0';
      elsif C_go = '1' then
         C_golatch <= '1';
      end if;
   end process;
   process (D_go, D_ack, D_flush) is
   begin
      if (D_ack = '1') or (D_flush = '1') then
         D_golatch <= '0';
      elsif D_go = '1' then
         D_golatch <= '1';
      end if;
   end process;

   -- mux first and last
   process (qsel, A_first, A_last, A_flush, B_first, B_last, B_flush,
            C_first, C_last, C_flush, D_first, D_last, D_flush) is
   begin
      case qsel is
         when B"00" =>
            first <= A_first;
            last <= A_last;
            flush <= A_flush;
         when B"01" =>
            first <= B_first;
            last <= B_last;
            flush <= B_flush;
         when B"10" =>
            first <= C_first;
            last <= C_last;
            flush <= C_flush;
         when others =>
            first <= D_first;
            last <= D_last;
            flush <= D_flush;
      end case;
   end process;
   -- compute veto
   middle <= not (first or last);
   veto <= '1' when ((first = '1') and (MC(13) = '1')) or
                    ((middle = '1') and (MC(14) = '1')) or
                    ((last = '1') and (MC(15) = '1')) or
                    (state = IDLE) else '0';

   -- reached end of program
   eofprog <= '1' when (MC(31) = '1') or (addr = B"1111111") or
                       (flush = '1') else '0';

   -- decoder finite state machine
   decoder_FSM: process (clk, longreset, state, A_golatch, B_golatch,
                         C_golatch, D_golatch, eofprog) is
   begin
      if rising_edge (clk) then
         if longreset = '1' then
            state <= IDLE;
            qsel <= B"00";
            addr <= (others => '0');
            A_ack <= '1';
            B_ack <= '1';
            C_ack <= '1';
            D_ack <= '1';
         else
            case state is
               when IDLE =>
                  addr <= (others => '0');
                  A_ack <= '0';
                  B_ack <= '0';
                  C_ack <= '0';
                  D_ack <= '0';
                  if A_golatch = '1' then
                     qsel <= B"00";
                     A_ack <= '1';
                     state <= GO;
                  elsif B_golatch = '1' then
                     qsel <= B"01";
                     B_ack <= '1';
                     state <= GO;
                  elsif C_golatch = '1' then
                     qsel <= B"10";
                     C_ack <= '1';
                     state <= GO;
                  elsif D_golatch = '1' then
                     qsel <= B"11";
                     D_ack <= '1';
                     state <= GO;
                  end if;
               when GO =>
                  A_ack <= '0';
                  B_ack <= '0';
                  C_ack <= '0';
                  D_ack <= '0';
                  if eofprog = '1' then
                     addr <= (others => '0');
                     state <= IDLE;
                  else
                     addr <= addr + 1;
                  end if;
            end case;
         end if;
      end if;
   end process decoder_FSM;

   -- queue select and microcode address
   SEL <= qsel;
   A <= std_logic_vector (addr);

   -- transmit, enable and end of frame marker
   enable <= '0' when veto = '1' else MC(16);
   transmit <= '0' when veto = '1' else MC(17);
   eof <= '0' when veto = '1' else eofprog;

   -- status
   decode_err <= '1' when (state = GO) and ((flush = '1') or
                ((addr = B"1111111") and (MC(31) = '0'))) else '0';
   status_1 <= B"000000000000000" & decode_err;
   status_2 <= status_1 when rising_edge (clk);
   status <= status_1 and not status_2;

end Structure;
--------------------------------------------------------------------------------
